import stomp
from tlib.base.LogHelper import get_logger
from uuid import uuid4


class ActiveMQListener(object):

    _id = uuid4().hex

    def __init__(self, name, headers=None, **keyword_headers):
        self.logger = get_logger(name="ActiveMQ listener")
        self._name = name
        self._headers = headers if headers else {}
        self._keyword_headers = keyword_headers

        self.logger.debug("""Initialized ActiveMQListener with this parameters:
    Name: {name}
    Id: {id}
    headers: {headers}
    keyword_headers: {keyword_headers}""".format(name=name, id=self._id, headers=headers,
                                                 keyword_headers=keyword_headers))

    def on_error(self, headers, message):
        self.logger.error("Method on_error must be implemented by classes receiving messages")
        raise NotImplementedError("Method on_error must be implemented by classes receiving messages")

    def on_message(self, headers, message):
        self.logger.error("Method on_message must be implemented by classes receiving messages")
        raise NotImplementedError("Method on_message must be implemented by classes receiving messages")

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    @property
    def headers(self):
        return self._headers

    @property
    def keyword_headers(self):
        return self._keyword_headers


class ActiveMQConnector(object):
    _logger = None

    def __init__(self, host, port, default_queue=None):
        self._logger = get_logger(name="ActiveMQ helper")
        self._default_queue = default_queue
        self._conn = stomp.Connection([(host, int(port))])

        self._logger.debug("Will connect to ActiveMQ on {host}:{port}".format(host=host, port=port))
        if default_queue:
            self._logger.debug("Using '{queue} as default".format(queue=default_queue))
            
    def connect(self):
        """
        Get configuration and connect to ActiveMQ
        """
        if not self._conn.is_connected():
            self._logger.debug("Connecting to ActiveMQ")
            self._conn.start()
            self._conn.connect()
        else:
            self._logger.warn("Already connected to ActiveMQ")

    def disconnect(self):
        """
        If there is a connection with ActiveMQ, close it.
        """
        if self._conn.is_connected():
            self._logger.debug("Disconnecting from ActiveMQ")
            # self._conn.stop()
            self._conn.disconnect()
        else:
            self._logger.warn("Already disconnected from ActiveMQ")
    
    def subscribe(self, listener, queue):
        """
        Adds a listener with the specified name

        :param listener: Listener that will receive messages
        :type listener: ActiveMQListener
        :param queue: Queue to subscribe to
        :type queue: str
        :raises: ValueError
        """
        #Input validation
        if not listener:
            self._logger.error("Listener was not provided")
            raise ValueError("Listener was not provided")

        if self._conn.get_listener(listener.name):
            self._logger.error("Listener name must be unique: %s" % listener.name)
            raise RuntimeError("Listener name must be unique: %s" % listener.name)

        #We need to be connected to subscribe
        self.connect()

        self._logger.debug("Registering listener '{name}'".format(name=listener.name))
        self._conn.set_listener(listener.name, listener)
        self._conn.subscribe(destination=queue, id=listener.id, headers=listener.headers, **listener.keyword_headers)

    def unsubscribe(self, listener):
        """
        Removes a listener

        :param listener: Listener that will receive messages
        :type listener: ActiveMQListener
        """
        #Input validation
        if not listener:
            self._logger.error("Listener was not provided")
            ValueError("Listener was not provided")

        if self._conn.is_connected():
            self._logger.debug("Unsubscribing listener '{name}'".format(name=listener.name))
            if self._conn.get_listener(listener.name):
                self._conn.unsubscribe(listener.id, headers=listener.headers, **listener.keyword_headers)
                self._conn.remove_listener(listener.name)
            else:
                self._logger.error("Listener with name {name} is not registered".format(name=listener.name))
                raise RuntimeError("Listener with name {name} is not registered".format(name=listener.name))
        else:
            self._logger.warn("Not connected to ActiveMQ. Can't unsubscribing")

    def send_message(self, message, queue=None, headers=None):
        """
        Send message to ActiveMQ.
        Id queue is not provided, it uses default queue

        :param: message: queue body
        :param: queue_destination: queue name
        """
        if headers is None:
            headers = {}

        #We need to be connected to send messages
        self.connect()

        if queue:
            self._logger.debug("Sending message to queue '{queue}'".format(queue=queue))
            self._conn.send(body=message, destination=queue, headers=headers)
        elif self._default_queue:
            self._logger.debug("Sending message to queue '{queue}'".format(queue=self._default_queue))
            self._conn.send(body=message, destination=self._default_queue, headers=headers)
        else:
            self._logger.error("Queue was not provided and there is no default queue")
            raise RuntimeError("Queue was not provided and there is no default queue")
